import time
import logging
import requests

from requests.packages import urllib3


class Stream(object):
    def __init__(self, host, url, listener, stream, token, protocol='https'):
        self.protocol = protocol
        self.host = host
        self.url = url
        self.stream = stream
        self.token = token
        self.listener = listener

        self.streaming = False
        self.retry_count = 0

    def go(self):
        url = '%s://%s%s' % (self.protocol, self.host, self.url)

        while True:
            try:
                self._go(url)
            except (requests.exceptions.ConnectionError, urllib3.exceptions.ProtocolError):
                logging.error('Connection failed unexpectedly with error.', exc_info=True)

            self._retry_wait()

    def _go(self, url):
        response = requests.get(
            url,
            {'stream': self.stream},
            stream=True,
            headers={
                'Authorization': 'Token %s' % self.token
            }
        )

        self._process_initial_response(response)
        self._consume(response)

    def _consume(self, response):
        buf = response.raw

        while not buf.closed:
            line = buf.readline().strip()

            if not line:
                self.listener.on_heartbeat()
            elif line.isdigit():
                length = int(line)
                header = buf.read(length)
                length = int(buf.readline().strip())
                body = buf.read(length - 2)
                self.listener.on_message(header, body)

        logging.warn('Connection closed, final message: %s', buf.read())

    def _process_initial_response(self, response):
        response.raise_for_status()

        logging.debug('Stream connection established.')
        self.streaming = True
        self.retry_count = 0

    def _retry_wait(self):
        self.retry_count += 1
        retry_wait = min(max((self.retry_count - 2), 0) ** 2, 60)  # Increase quadratically to a maximum of 60s.

        logging.warning('Attempting reconnect number %d in %d seconds.', self.retry_count, retry_wait)

        time.sleep(retry_wait)



import time

from .utils import concurrent_get

MAX_PENDING = 10000


class ResultRetrieved(Exception):
    pass


class InvalidTaskStatus(Exception):
    pass


class ObjectStore(object):
    def __init__(self, api):
        self.api = api

    def go(self):
        while True:
            time.sleep(1)
            pending = self._pending_tasks()
            if pending:
                self._handle_pending_tasks(pending)
                time.sleep(5)

    def _pending_tasks(self):
        return self.api.pending_tasks.keys() if self.api.pending_tasks else []

    @staticmethod
    def _check_status_integrity(status):
        # Check if everything is ok with the status
        if 'result_retrieved' not in status or 'status' not in status:
            raise InvalidTaskStatus
        if status['result_retrieved']:
            raise ResultRetrieved

    def _get_task_status(self, pending):
        return self.api.task_status(pending[:MAX_PENDING])

    def _set_result_in_memory(self, uuid, result):
        # Make the result accessible to the other thread
        self.api.set_task_result(uuid, result)

    def _handle_pending_tasks(self, pending):
        task_status_response = self._get_task_status(pending)
        task_status_response.pop('success')

        retrieval_urls = []
        for task_id, task_status in task_status_response.iteritems():
            if 'retrieval_endpoint' in task_status:
                url = (task_id, task_status['retrieval_endpoint'])
                retrieval_urls.append(url)

        if retrieval_urls:
            results = self._retrieve_result(retrieval_urls, self.api.token_header)
            for task_id, task_result in results.iteritems():
                self._set_result_in_memory(task_id, task_result)

    @staticmethod
    def _retrieve_result(endpoints, token_header):
        """Prepare the request list and execute them concurrently."""
        request_list = [
            (url, token_header)
            for (task_id, url) in endpoints
        ]

        responses = concurrent_get(request_list)

        # Quick sanity check
        assert len(endpoints) == len(responses)

        responses_dic = {
            task_id: r.content
            for (task_id, _), r in zip(endpoints, responses)
        }
        return responses_dic



import os
import ConfigParser


def get_config(config_name='ricloud.ini'):

    config = ConfigParser.RawConfigParser()
    path_to_config = os.path.join(os.path.dirname(__file__), config_name)

    home_path = os.path.expanduser(os.path.join('~', '.%s' % config_name))
    paths = [path_to_config, home_path]

    if 'RICLOUD_CONF' in os.environ:
        paths.append(os.environ['RICLOUD_CONF'])

    config.read(paths)

    return config


settings = get_config()

BASE_DIR = os.getcwd()

OUTPUT_DIR = os.path.join(BASE_DIR, settings.get('output', 'output_directory'))

LOG_DIRECTORY = os.path.join(BASE_DIR, settings.get('logging', 'logs_directory'))

DEFAULT_LOG = os.path.join(LOG_DIRECTORY, 'ricloud.log')

LOGGING_LEVEL = settings.get('logging', 'level')

LISTENER_DB_HOST = settings.get('mysql', 'host')
LISTENER_DB_PORT = settings.get('mysql', 'port')
LISTENER_DB_NAME = settings.get('mysql', 'database')
LISTENER_DB_USER = settings.get('mysql', 'username')
LISTENER_DB_PASSWORD = settings.get('mysql', 'password')



"""ricloud API Client.

Usage:
    ricloud <account> [--password=<password>] [--timeout=<timeout>]
    ricloud --listen [--timeout=<timeout>]
    ricloud --list-services [--timeout=<timeout>]
    ricloud --list-subscriptions <service> [--timeout=<timeout>]
    ricloud --subscribe-account <username> <password> <service> [--timeout=<timeout>]
    ricloud --perform-2fa-challenge <account_id> <device_id> [--timeout=<timeout>]
    ricloud --submit-2fa-challenge <account_id> <code> [--timeout=<timeout>]
    ricloud --resubscribe-account <account_id> <password> [--timeout=<timeout>]
    ricloud --unsubscribe-account <account_id> [--timeout=<timeout>]
    ricloud --list-devices <account_id> [--timeout=<timeout>]
    ricloud --subscribe-device <account_id> <device_id> [--timeout=<timeout>]
    ricloud --unsubscribe-device <account_id> <device_id> [--timeout=<timeout>]
    ricloud --reset-subscription-since <account_id> <datetime> [--timeout=<timeout>]
    ricloud --help


Information:
    ricloud can be run in three modes:

    Listen mode:
        Starts a process which listends to feeds, system messages, and files being sent from asmaster, and saves these in the ricloud database.
        ricloud --listen [--timeout=<timeout>]

    Manager mode:
        To send messages to asmaster, such as to register a new account.
        The following actions are available with this mode:
            --list-services --list-subscriptions, --subscribe-account, --perform-2fa-challenge, --submit-2fa-challenge, --resubscribe-account, --unsubscribe-account, --list-devices, --subscribe-device, --unsubscribe-device

    Legacy interactive mode:
        For logging in to and downloading data from an account using with, without registering the account with asmaster (and without the corresponding locking prevention).
        ricloud <account> [--password=<password>] [--timeout=<timeout>]


Options:
    -h --help               Show this screen.
    --password=<password>   The password for this account.
    --timeout=<timeout>     How long should we wait for tasks to complete, in seconds. (default: 600s or 10 minutes)

"""


from __future__ import unicode_literals

import logging
import json
import os
from docopt import docopt

from .ricloud import RiCloud
from .conf import DEFAULT_LOG, LOGGING_LEVEL, LOG_DIRECTORY
from .asmaster_api import AsmasterApi
from .utils import select_service, select_samples


def _parse_input_arguments(arguments):
    timeout = int(arguments.get('--timeout') or 600)

    if arguments['--listen']:
        return {
            'mode': 'listen',
            'timeout': timeout,
        }

    manager_mode_actions = [
        'list-services',
        'list-subscriptions',
        'subscribe-account',
        'perform-2fa-challenge',
        'submit-2fa-challenge',
        'resubscribe-account',
        'unsubscribe-account',
        'list-devices',
        'subscribe-device',
        'unsubscribe-device',
        'reset-subscription-since',
    ]

    for action in manager_mode_actions:
        if arguments["--{}".format(action)]:
            return {
                'mode': 'manager',
                'timeout': timeout,
            }

    account = arguments['<account>']
    password = arguments['--password']

    return {
        'mode': 'interactive',
        'account': account,
        'password': password,
        'timeout': timeout,
    }


def main():
    # Get input arguments from command line
    arguments = docopt(__doc__)

    # Format the input arguments for the application
    application_payload = _parse_input_arguments(arguments)

    if not os.path.exists( LOG_DIRECTORY ):
        os.makedirs( LOG_DIRECTORY )

    # setup the logger
    logging.basicConfig(filename=DEFAULT_LOG,
                        filemode='a',
                        format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                        level=LOGGING_LEVEL)

    if application_payload['mode'] == 'interactive':
        # Main RiCloud object
        ricloud = RiCloud()

        # The RiCloud client retrieves a list of available services from the API.
        service_name = select_service(ricloud)

        # Each service has a client which contains all the available tasks.
        # Sample applications for these clients can be found in the `samples` folder.
        Application = select_samples(ricloud, service_name, application_payload)

        # Execute the application's main function.
        Application.run()

    elif application_payload['mode'] == 'listen':
        # asmaster listener requires MySQLdb, and we want to avoid requiring that otherwise
        from .asmaster_listener import AsmasterListener

        # This sets the listener running and will not return
        AsmasterListener(application_payload['timeout'])

    elif application_payload['mode'] == 'manager':
        api = AsmasterApi(application_payload['timeout'])
        api.setup()

        if(arguments["--list-services"]):
            print json.dumps(api.list_services(), indent=2)
        if(arguments["--list-subscriptions"]):
            print json.dumps(api.list_subscriptions(arguments['<service>']), indent=2)
        if(arguments["--subscribe-account"]):
            print json.dumps(api.subscribe_account(
                arguments['<username>'],
                arguments['<password>'],
                arguments['<service>']), indent=2)
        if(arguments["--perform-2fa-challenge"]):
            print json.dumps(api.perform_2fa_challenge(arguments['<account_id>'], arguments['<device_id>']), indent=2)
        if(arguments["--submit-2fa-challenge"]):
            print json.dumps(api.submit_2fa_challenge(arguments['<account_id>'], arguments['<code>']), indent=2)
        if(arguments["--resubscribe-account"]):
            print json.dumps(api.resubscribe_account(arguments['<account_id>'], arguments['<password>']), indent=2)
        if(arguments["--unsubscribe-account"]):
            print json.dumps(api.unsubscribe_account(arguments['<account_id>']), indent=2)
        if(arguments["--list-devices"]):
            print json.dumps(api.list_devices(arguments['<account_id>']), indent=2)
        if(arguments["--subscribe-device"]):
            print json.dumps(api.subscribe_device(arguments['<account_id>'], arguments['<device_id>']), indent=2)
        if(arguments["--unsubscribe-device"]):
            print json.dumps(api.unsubscribe_device(arguments['<account_id>'], arguments['<device_id>']), indent=2)
        if(arguments["--reset-subscription-since"]):
            print json.dumps(api.reset_subscription_since(arguments['<account_id>'], arguments['<datetime>']), indent=2)


if __name__ == '__main__':
    main()



import os
import re
import json
import shutil
import logging
import hashlib
import MySQLdb

from .api import Api
from .ricloud import RiCloud
from .listener import Listener
from .handlers import RiCloudHandler, StreamError
from . import conf


class AsmasterListener(RiCloud):
    def __init__(self, timeout, api=None):
        self.timeout = timeout

        self.api = api or Api()

        super(AsmasterListener, self).__init__(
            api=self.api,
            listener=Listener({'__ALL__': DatabaseWritingHandler(api=self.api)})
        )

    @property
    def stream_thread_is_daemon(self):
        return False


class DatabaseWritingHandler(RiCloudHandler):
    def __init__(self, api=None, callback=None, db=conf.LISTENER_DB_NAME, file_location=conf.OUTPUT_DIR):
        self.db = db
        self.file_location = file_location
        super(DatabaseWritingHandler, self).__init__(api, callback)

    _db_con = None

    @property
    def db_con(self):
        if not self._db_con:
            self._db_con = MySQLdb.connect(
                host=conf.LISTENER_DB_HOST,
                port=int(conf.LISTENER_DB_PORT),
                user=conf.LISTENER_DB_USER,
                passwd=conf.LISTENER_DB_PASSWORD,
                db=self.db
            )
        return self._db_con

    def handle_query(self, query, args=None, retry=2):
        try:
            cursor = self.db_con.cursor()
            cursor.execute(query, args=args)
            self.db_con.commit()
        except (AttributeError, MySQLdb.OperationalError):
            if not retry:
                raise

            logging.warn('asmaster listener handler query failed, attempting to refresh database connection.')
            self._db_con = None
            self.handle_query(query, args=args, retry=retry - 1)

    def on_complete_message(self, header, stream):
        if header['type'] == 'system':
            body = stream.read()
            json_body = json.loads(body)

            try:
                code = json_body['code'][:200]
            except KeyError:
                code = None

            query = """
                INSERT INTO system (`received`, `headers`, `body`, `message`, `code`)
                VALUES (NOW(), %(headers)s, %(body)s, %(message)s, %(code)s)
            """

            args = {
                "headers": json.dumps(header),
                "body": body,
                "message": json_body['message'][:200],
                "code": code,
            }

            self.handle_query(query, args)

            if "error" in body:
                raise StreamError(body)

        elif header['type'] in {'fetch-data', 'message'}:
            if header['type'] == 'fetch-data':
                table = 'feed'
            else:
                table = header['type']

            query = """
                INSERT INTO {table} (`service`, `received`, `account_id`, `device_id`, `device_tag`, `headers`, `body`)
                VALUES (%(service)s, NOW(), %(account_id)s, %(device_id)s, %(device_tag)s, %(headers)s, %(body)s)
            """.format(table=table)  # noqa

            args = {
                "table": table,
                "service": header['service'],
                "account_id": header.get('account_id', None),
                "device_id": header.get('device_id', None),
                "device_tag": header.get('device_tag', None),
                "headers": json.dumps(header),
                "body": json.dumps(stream.read()),
            }

            self.handle_query(query, args)
            self.api.result_consumed(header.get('task_id'))

        elif header['type'] == 'download-file':
            try:
                file_id = header['file_id']
            except KeyError:
                raise StreamError("Invalid download file request, no file_id")
            if len(file_id) > 4096:
                raise StreamError("Invalid download file request, file_id is too long")

            file_path = self.save_stream_to_file(header, stream, self.file_location)

            query = """
                INSERT INTO file (`service`, `received`, `account_id`, `device_id`, `device_tag`, `headers`, `location`, `file_id`)
                VALUES (%(service)s, NOW(), %(account_id)s, %(device_id)s, %(device_tag)s, %(headers)s, %(location)s, %(file_id)s)
            """  # noqa

            args = {
                "service": header['service'],
                "account_id": header.get('account_id', None),
                "device_id": header.get('device_id', None),
                "device_tag": header.get('device_tag', None),
                "headers": json.dumps(header),
                "location": file_path,
                "file_id": file_id,
            }

            self.handle_query(query, args)
            self.api.result_consumed(header.get('task_id'))

        else:
            raise StreamError("Unrecognised header type {}".format(header['type']))

    @classmethod
    def save_stream_to_file(cls, header, stream, file_location):
        filename = cls.file_id_to_file_name(header['file_id'])
        path = os.path.join(
            file_location,
            header['service'],
            "{}".format(header.get('account_id', "None")),
            str(header.get('device_id', "None")),
            filename,
        )

        if len(path) > 250:
            raise StreamError("File path too long, unable to save stream.")

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        if os.path.isfile(path):
            os.remove(path)

        with open(path, 'w') as f:
            shutil.copyfileobj(stream, f)

        return path

    @staticmethod
    def file_id_to_file_name(file_id):
        """
        Sometimes file ids are not the file names on the device, but are instead generated
        by the API. These are not guaranteed to be valid file names so need hashing.
        """
        if len(file_id) == 40 and re.match("^[a-f0-9]+$", file_id):
            return file_id
        # prefix with "re_" to avoid name collision with real fileids
        return "re_{}".format(hashlib.sha1(file_id).hexdigest())



from __future__ import print_function

import os
import sys
import time
import json
import datetime
import requests
from multiprocessing.pool import ThreadPool

from clint.textui import prompt, puts, colored, indent

from .conf import settings, OUTPUT_DIR
from .samples import get_samples


def get_or_create_filepath(filename, directory=''):
    absolute_dir = os.path.join(OUTPUT_DIR, directory)

    if not os.path.exists(absolute_dir):
        os.makedirs(absolute_dir)

    return os.path.join(absolute_dir, filename)


def utf8(message):
    if isinstance(message, unicode):
        message = message.encode('utf-8')
    return message


def print_message(message, colour='white'):
    """Prints `message` to the console in the specified `colour`.
    Makes sure special characters are encoded properly (for example, emoticons
    in device names).
    """
    puts(getattr(colored, colour)(utf8(message)))


def info_message(message):
    """Prints info type messages in green."""
    print_message(message, colour='green')


def prompt_message(message):
    """Prints prompt type messages in blue."""
    print_message('\n' + message, colour='blue')


def pending_message(message):
    """Prints pending type messages in yellow."""
    print_message('\n' + message, colour='yellow')


def error_message(message):
    """Prints error type messages in red."""
    print_message('\n' + message, colour='red')


def error_message_and_exit(message, error_result):
    """Prints error messages in blue, the failed task result and quits."""
    if message:
        error_message(message)
    puts(json.dumps(error_result, indent=2))
    sys.exit(1)


def print_prompt_values(values, message=None, sub_attr=None):
    """Prints prompt title and choices with a bit of formatting."""
    if message:
        prompt_message(message)

    for index, entry in enumerate(values):
        if sub_attr:
            line = '{:2d}: {}'.format(index, getattr(utf8(entry), sub_attr))
        else:
            line = '{:2d}: {}'.format(index, utf8(entry))

        with indent(3):
            print_message(line)


def prompt_for_input(message, input_type=None):
    """Prints prompt instruction and does basic input parsing."""
    while True:
        output = prompt.query(message)

        if input_type:
            try:
                output = input_type(output)
            except ValueError:
                error_message('Invalid input type')
                continue

        break

    return output


def prompt_for_choice(values, message, input_type=int, output_type=None):
    """Prints prompt with a list of choices to choose from."""
    output = None
    while not output:
        index = prompt_for_input(message, input_type=input_type)

        try:
            output = utf8(values[index])
        except IndexError:
            error_message('Selection out of range')
            continue

    if output_type:
        output = output_type(output)

    return output


def select_item(items, prompt_instruction, prompt_title, sub_attr=None, output_type=None):
    print_prompt_values(items, prompt_title, sub_attr)

    return prompt_for_choice(items, prompt_instruction, output_type=output_type)


def select_service(response):
    return select_item(
        response.services,
        'Please select a service index:',
        'Authorized services:',
        output_type=str
    )


def select_samples(response, service_name, payload):
    samples = get_samples(service_name)

    selected_sample = select_item(
        samples,
        'Please select a sample application index:',
        'Available sample applications:',
        sub_attr='display_name'
    )

    return selected_sample(response, payload)


def profile(timer_text=''):
    def inner(func):
        def wraps(*args, **kwargs):
            timer_start = time.time()
            ret = func(*args, **kwargs)
            timer_end = time.time()
            if settings.getboolean('logging', 'time_profile'):
                delta = timer_end - timer_start
                message = "{text} {delta:.2f}s".format(text=timer_text, delta=delta)
                puts(colored.magenta(message))
            return ret
        return wraps
    return inner


def append_profile_info(string, info):
    return utf8(string) + colored.magenta(" completed in {0:.2f}s".format(info))


def _get_num_threads():
    num_threads = int(settings.get('performance', 'object_store_greenlets'))
    if not num_threads:
        num_threads = 1
    return num_threads


def threaded_get(request_instance):
    return requests.request(
        method='get',
        url=request_instance[0],
        headers=request_instance[1],
    )


def concurrent_get(request_list):
    max_threads = _get_num_threads()
    number_requests = len(request_list)
    if number_requests > 0:
        pool = ThreadPool(number_requests) if number_requests <= max_threads else ThreadPool(max_threads)
        results = pool.map(threaded_get, request_list)
        pool.close()
        pool.join()
        return results
    else:
        return []



import json

from .icloud_sample import SampleICloudApplication

from .. import utils


class SampleLiveICloudApplication(SampleICloudApplication):
    display_name = 'Sample Live Data Application'

    def run(self):
        # Register the account for the iCloud service.
        self.client.register_account(self.account)
        # Attempt to login to the account.
        self.log_in()
        # Choose a data type and retrieve it.
        self.fetch_data()
        # Wait for any pending tasks to complete
        self.client.wait_for_pending_tasks()
        utils.info_message('All tasks completed')

    def fetch_data(self):
        """Prompt for a data type choice and execute the `fetch_data` task.
        The results are saved to a file in json format.
        """
        choices = self.available_data
        choices.insert(0, 'All')

        selected_data_type = utils.select_item(
            choices,
            'Please select what data to fetch:',
            'Available data:',
        )

        if selected_data_type == 'All':
            selected_data_type = ','.join(self.available_data)

        utils.pending_message('Performing fetch data task...')

        fetch_data_task = self.client.data(
            account=self.account,
            data=selected_data_type,
        )

        # Wait here for result as rest of sample app relies on it.
        fetch_data_task.wait_for_result(timeout=self.timeout)
        fetch_data_result = json.loads(fetch_data_task.result)

        # Write the result to file.
        task_id = fetch_data_task.uuid
        filepath = utils.get_or_create_filepath('%s.json' % task_id)

        with open(filepath, 'w') as out:
            json.dump(fetch_data_result, out, indent=2)

        utils.info_message('Fetch data successful. Output file: %s.json' % task_id)

        return fetch_data_result

    @property
    def available_data(self):
        live_feeds = [
            'location',
            'mobileme_contacts',
            'web_browser_history',
            'live_call_history',
            'live_photos',
        ]
        return [c for c in self.client.available_data
                if c in live_feeds]



from __future__ import print_function

import json
import logging
import traceback

from .handlers import StreamError


class Listener(object):
    """Handle stream messagse here."""
    def __init__(self, handlers):
        self.handlers = handlers

    def on_heartbeat(self):
        logging.debug('-*- Heartbeat -*-')

    def on_message(self, header, data):
        header = json.loads(header)
        handler = self.handlers.get(header['type'], self.handlers.get('__ALL__', None))

        if not handler:
            logging.error('Unknown message type', header['type'])
            logging.error(header)
            logging.error(data)
        else:
            try:
                handler.handle(header, data)
            except StreamError as e:
                logging.error("StreamError: %s", e.message)
            except Exception as e:
                logging.critical("%s: %s", type(e), e.args)
                logging.critical(traceback.format_exc())
